import logging
import click
import requests
from os.path import basename

from multiprocessing import Pool, current_process

from geoseeq.cli.constants import *
from geoseeq.cli.shared_params import (
    handle_project_id,
    private_option,
    link_option,
    module_option,
    project_id_arg,
    overwrite_option,
    yes_option,
    use_common_state,
)

from geoseeq.constants import FASTQ_MODULE_NAMES
from geoseeq.cli.progress_bar import PBarManager

logger = logging.getLogger('geoseeq_api')


def _make_in_process_logger(log_level):
    logger = logging.getLogger('geoseeq_api')
    logger.setLevel(log_level)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('[%(levelname)s] %(name)s :: ' + current_process().name + ' :: %(message)s'))
    logger.addHandler(handler)
    return logger


def _upload_one_file(args):
    result_file, filepath, session, progress_tracker, link_type, overwrite, log_level = args
    _make_in_process_logger(log_level)
    if result_file.exists() and not overwrite:  
        return
    result_file = result_file.idem()
    if link_type == 'upload':
        # TODO: check checksums to see if the file is the same
        result_file.upload_file(filepath, session=session, progress_tracker=progress_tracker, threads=4)
    else:
        result_file.link_file(link_type, filepath)


def _get_regex(knex, filepaths, module_name, lib, regex):
    """Return a regex that will group the files into samples
    
    Tell the user how many files did could not be matched using the regex.
    """
    seq_length, seq_type = module_name.split('::')[:2]
    args = {
        'filenames': list(filepaths.keys()),
        'sequence_type': seq_type,
        'sample_group_id': lib.uuid,
    }
    if regex:
        args['custom_regex'] = regex
    result = knex.post('bulk_upload/validate_filenames', json=args)
    regex = result['regex_used']
    click.echo(f'Using regex: "{regex}"', err=True)
    if result['unmatched']:
        click.echo(f'{len(result["unmatched"])} files could not be grouped.', err=True)
    else:
        click.echo('All files successfully grouped.', err=True)
    return regex


def _group_files(knex, filepaths, module_name, regex, yes):
    """Group the files into samples, confirm, and return the groups."""
    seq_length, seq_type = module_name.split('::')[:2]
    groups = knex.post('bulk_upload/group_files', json={
        'filenames': list(filepaths.keys()),
        'sequence_type': seq_type,
        'regex': regex
    })
    for group in groups:
        click.echo(f'sample_name: {group["sample_name"]}', err=True)
        click.echo(f'  module_name: {module_name}', err=True)
        for field_name, filename in group['fields'].items():
            path = filepaths[filename]
            click.echo(f'    {seq_length}::{field_name}: {path}', err=True)
    if not yes:
        click.confirm('Do you want to upload these files?', abort=True)
    return groups


def _do_upload(groups, module_name, link_type, lib, filepaths, overwrite, cores, state):
    def handle_upload(sample, success, error):
        if success:
            click.echo(f'Uploaded Sample: {sample.name}', err=True)
        else:
            click.echo(f'Failed to upload Sample: {sample.name}', err=True)
            click.echo(f'Error:\n{error}\n{error.traceback()}', err=True)

    pbars = PBarManager()
    with requests.Session() as session:
        upload_args = []
        for group in groups:
            sample = lib.sample(group['sample_name']).idem()
            read_folder = sample.result_folder(module_name).idem()

            for field_name, path in group['fields'].items():
                result_file = read_folder.read_file(field_name)
                filepath = filepaths[path]
                upload_args.append((
                    result_file, filepath, session, pbars.get_new_bar(filepath),
                    link_type, overwrite, state.log_level
                ))

        with Pool(cores) as p:
            for _ in p.imap_unordered(_upload_one_file, upload_args):
                pass


def _is_fastq(path, fq_exts=['.fastq', '.fq'], compression_exts=['.gz', '.bz2', '']):
    for fq_ext in fq_exts:
        for compression_ext in compression_exts:
            if path.endswith(fq_ext + compression_ext):
                return True
    return False


def flatten_list_of_fastqs(filepaths):
    """Turn a list of fastq filepaths and txt files containing fastq filepaths into a single list of fastq filepaths."""
    flattened = []
    for path in filepaths:
        if _is_fastq(path):
            flattened.append(path)
        else:
            with open(path) as f:
                for line in f:
                    line = line.strip()
                    if line and not line.startswith('#'):
                        flattened.append(line)
    return flattened



@click.command('reads')
@use_common_state
@click.option('--cores', default=1, help='Number of uploads to run in parallel')
@overwrite_option
@yes_option
@click.option('--regex', default=None, help='An optional regex to use to extract sample names from the file names')
@private_option
@link_option
@module_option(FASTQ_MODULE_NAMES)
@project_id_arg
@click.argument('fastq_files', type=click.Path(exists=True), nargs=-1)
def cli_upload_reads_wizard(state, cores, overwrite, yes, regex, private, link_type, module_name, project_id, fastq_files):
    """Upload fastq read files to GeoSeeq.

    This command automatically groups files by their sample name, lane number
    and read number. It asks for confirmation before creating any samples or
    data.

    ---

    Example Usage:

    \b
    # Upload a list of fastq files to a project, useful if you have hundreds of files
    $ ls -1 path/to/fastq/files/*.fastq.gz > file_list.txt
    $ geoseeq upload reads "GeoSeeq/Example CLI Project" file_list.txt

    \b
    # Upload all the fastq files in a directory to a project
    $ geoseeq upload reads ed59b913-91ec-489b-a1b9-4ea137a6e5cf path/to/fastq/files/*.fastq.gz

    \b
    # Upload all the fastq files in a directory to a project, performing 4 uploads in parallel
    $ geoseeq upload reads --cores 4 ed59b913-91ec-489b-a1b9-4ea137a6e5cf path/to/fastq/files/*.fastq.gz

    \b
    # Upload a list of fastq files to a project, automatically creating a new project and overwriting existing files
    $ ls -1 path/to/fastq/files/*.fastq.gz > file_list.txt
    $ geoseeq upload reads --yes --overwrite "GeoSeeq/Example CLI Project" file_list.txt

    ---

    Command Arguments:
    
    [PROJECT_ID] Can be a project UUID, GeoSeeq Resource Number (GRN), or an
    organization name and project name separated by a slash.

    \b
    Examples: 
     - Name pair: "GeoSeeq/Example CLI Project"
     - UUID: "ed59b913-91ec-489b-a1b9-4ea137a6e5cf"
     - GRN: "grn:gs1:project:ed59b913-91ec-489b-a1b9-4ea137a6e5cf"

    \b
    [FASTQ_FILES...] can be paths to fastq files or a file containing a list of paths, or a mix of both.
    Example: "path/to/fastq/files/*.fastq.gz" "file_list.txt" "path/to/more/fastq/files/*.fastq.gz"

    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, yes, private)
    filepaths = {basename(line): line for line in flatten_list_of_fastqs(fastq_files)}
    click.echo(f'Found {len(filepaths)} files to upload.', err=True)
    regex = _get_regex(knex, filepaths, module_name, proj, regex)
    groups = _group_files(knex, filepaths, module_name, regex, yes)
    _do_upload(groups, module_name, link_type, proj, filepaths, overwrite, cores, state)



import uuid


def is_grn(el):
    """Return True if `el` is a GeoSeeq Resource Number (GRN)"""
    return el.startswith('grn:')

def is_uuid(el):
    """Return True if `el` is a UUID"""
    try:
        uuid.UUID(el)
        return True
    except ValueError:
        return False

def is_grn_or_uuid(el):
    """Return True if `el` is a GRN or a UUID"""
    return is_grn(el) or is_uuid(el)




from .remote_object import RemoteObject
from .project import Project


class Organization(RemoteObject):
    remote_fields = [
        'uuid',
        'created_at',
        'updated_at',
        'name',
    ]
    parent_field = None
    url_prefix = 'organizations'

    def __init__(self, knex, name):
        super().__init__(self)
        self.knex = knex
        self.name = name

    def nested_url(self):
        return f'nested/{self.name}'

    def _save(self):
        data = {
            field: getattr(self, field)
            for field in self.remote_fields if hasattr(self, field)
        }
        url = f'{self.url_prefix}/{self.uuid}'
        self.knex.put(url, json=data)

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(self.nested_url())
            self.load_blob(blob, allow_overwrite=allow_overwrite)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def _create(self):
        blob = self.knex.post(self.url_prefix, json={'name': self.name})
        self.load_blob(blob)

    def sample_group(self, *args, **kwargs):
        """Create a new project in this organization.
        
        This is an alias for project() for backwards compatibility.
        """
        return self.project(*args, **kwargs)

    def project(self, project_name, metadata={}, is_public=False):
        """Create a new project in this organization."""
        return Project(self.knex, self, project_name, is_public=is_public, metadata=metadata)

    def get_projects(self):
        """Yield projects in this org fetched from the server."""
        url = f'sample_groups?organization_id={self.uuid}'
        result = self.knex.get(url)
        for result_blob in result['results']:
            result = self.sample_group(result_blob['name'])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            yield result

    def get_sample_groups(self):
        """Yield projects in this org fetched from the server.

        This is an alias for get_projects() for backwards compatibility.
        """
        return self.get_projects()


    def pre_hash(self):
        return 'ORG' + self.name



from os.path import join


class VCCache:
    """A filesystem cache so that multiple users
    on the same machine can avoid excess downloads and storage.
    """

    def __init__(self, root_path):
        self.root = root_path

    def get_cache_filepath(self, field):
        """Return a filepath in the cache to be used for `field`.

         - `field ` must exist on the remote
         - Does not matter if the filepath exists or not.
        """
        field = field.get()
        checksum_blob = field.checksum()
        base = f'{field.uuid}.{checksum_blob["method"]}__{checksum_blob["value"]}'
        path = join(self.root, field.uuid[-2:], base)  # -2 in case uuids have a timestamp prefix
        return path
    





from .result_file import (
    ResultFile,
    AnalysisResultField,
    SampleResultFile,
    SampleAnalysisResultField,
    ProjectResultFile,
    SampleGroupAnalysisResultField,
)
from .result_folder import (
    ResultFolder,
    AnalysisResult,
    SampleResultFolder,
    SampleAnalysisResult,
    ProjectResultFolder,
    SampleGroupAnalysisResult,
)


import xml.etree.ElementTree as ET

from Bio import Entrez


def recurse_xml_to_reactjs_safe_dict(v):
    out = {"tag": v.tag, "attributes": [v.attrib], "children": []}
    for child in v:
        out["children"].append(recurse_xml_to_reactjs_safe_dict(child))
    return out


def print_xml_recurse(v, pref=""):
    print(pref, v.tag, v.attrib)
    for child in v:
        print_xml_recurse(child, pref + "\t")


def xml_sra_file_to_blob(sra):
    out = {
        "__type__": "sra",
        "url": sra.attrib["url"],
        "md5": sra.attrib["md5"],
        "full_record": sra.attrib,
        "alternatives": [child.attrib for child in sra.iter("Alternatives")],
    }
    return out


class BioProject:
    def __init__(self, accession):
        self.accession = accession
        self._sra_records = None
        self.root = None

    def fetch(self):
        if self.root:
            return self
        handle = Entrez.efetch(db="bioproject", id=self.accession_num)
        self.root = ET.fromstring(handle.read())
        return self

    def metadata(self):
        self.fetch()
        out = {"ncbi_xml": ET.tostring(self.root, encoding="unicode", method="xml")}
        return out

    @property
    def accession_num(self):
        return self.accession.split("PRJNA")[-1]

    def biosamples(self):
        handle = Entrez.elink(
            dbfrom="bioproject", id=self.accession_num, linkname="bioproject_biosample"
        )
        results = Entrez.read(handle)
        out = []
        for el in results[0]["LinkSetDb"][0]["Link"]:
            out.append(BioSample(el["Id"], self))
        return out

    def geoseeq_obj(self, org):
        self.fetch()
        grp = org.sample_group(self.accession, metadata=self.metadata(), is_library=True)
        return grp

    def __str__(self):
        return f"<BioProject accession={self.accession} />"


class BioSample:
    def __init__(self, accession, bioproject=None):
        self.bioproject = bioproject
        self.accession = accession
        self.root = None

    @property
    def accession_num(self):
        return self.accession.split("SAMN")[-1]

    def fetch(self):
        if self.root:
            return self
        handle = Entrez.efetch(db="biosample", id=self.accession_num)
        self.root = ET.fromstring(handle.read())
        biosample_tag = self.root[0]
        self.accession = biosample_tag.attrib["accession"]
        return self

    def metadata(self):
        self.fetch()
        out = {"ncbi_xml": ET.tostring(self.root, encoding="unicode", method="xml")}
        return out

    def sra(self):
        handle = Entrez.elink(dbfrom="biosample", id=self.accession_num, linkname="biosample_sra")
        results = Entrez.read(handle)
        out = []
        for el in results[0]["LinkSetDb"][0]["Link"]:
            out.append(SRARecord(el["Id"], self))
        return out

    def geoseeq_obj(self, grp):
        self.fetch()
        sample = grp.sample(self.accession, metadata=self.metadata())
        return sample

    def __str__(self):
        return f"<BioSample accession={self.accession} />"


class SRARecord:
    def __init__(self, accession, biosample=None):
        self.biosample = biosample
        self.accession = accession
        self.root = None

    def fetch(self):
        if self.root:
            return self
        handle = Entrez.efetch(db="sra", id=self.accession_num)
        self.root = ET.fromstring(handle.read())
        return self

    def metadata(self):
        self.fetch()
        out = {"ncbi_xml": ET.tostring(self.root, encoding="unicode", method="xml")}
        return out

    def analysis_result_blob(self):
        self.fetch()
        sra_files = [sra_file for sra_file in self.root.iter("SRAFile")]
        assert len(sra_files) == 1
        sra_file = sra_files[0]
        blob = xml_sra_file_to_blob(sra_file)
        return blob

    @property
    def accession_num(self):
        return self.accession.split("PRJNA")[-1]

    def geoseeq_obj(self, sample):
        self.fetch()
        ar = sample.analysis_result(
            "raw::raw_reads", replicate=self.accession, metadata=self.metadata()
        )
        field = ar.field("sra_run", data=self.analysis_result_blob())
        return ar, field

    def __str__(self):
        return f"<SRARecord accession={self.accession} />"




import urllib.request
import logging
import requests
from os.path import basename, getsize, join
from pathlib import Path
from tempfile import NamedTemporaryFile

from geoseeq.utils import download_ftp
from geoseeq.constants import FIVE_MB

logger = logging.getLogger("geoseeq_api")  # Same name as calling module


def _download_head(url, filename, head=None, progress_tracker=None):
    headers = None
    if head and head > 0:
        headers = {"Range": f"bytes=0-{head}"}
    response = requests.get(url, stream=True, headers=headers)
    response.raise_for_status()
    total_size_in_bytes = int(response.headers.get('content-length', 0))
    if progress_tracker: progress_tracker.set_num_chunks(total_size_in_bytes)
    block_size = FIVE_MB
    with open(filename, 'wb') as file:
        for data in response.iter_content(block_size):
            if progress_tracker: progress_tracker.update(len(data))
            file.write(data)
    return filename


def _download_generic(url, filename, head=None):
    urllib.request.urlretrieve(url, filename)
    return filename


def guess_download_kind(url):
    if 'azure' in url:
        return 'azure'
    elif 's3' in url:
        return 's3'
    elif 'ftp' in url:
        return 'ftp'
    else:
        return 'generic'


def download_url(url, kind='guess', filename=None, head=None, progress_tracker=None):
    """Return a local filepath to the downloaded file. Download the file."""
    if kind == 'guess':
        kind = guess_download_kind(url)
        logger.info(f"Guessed download kind: {kind} for {url}")
    logger.info(f"Downloading {kind} file to {filename}")
    if kind == 'generic':
        return _download_generic(url, filename, head=head)
    elif kind == 's3':
        return _download_head(url, filename, head=head, progress_tracker=progress_tracker)
    elif kind == 'azure':
        return _download_head(url, filename, head=head)
    elif kind == 'ftp':
        return download_ftp(url, filename, head=head)
    else:
        raise ValueError(f"Unknown download kind: {kind}")



class ResultFileDownload:
    """Abstract class that handles download methods for result files."""

    def get_download_url(self):
        """Return a URL that can be used to download the file for this result."""
        blob_type = self.stored_data.get("__type__", "").lower()
        if blob_type not in ["s3", "sra", "ftp", "azure"]:
            raise ValueError(f'Unknown URL type: "{blob_type}"')
        key = 'url' if 'url' in self.stored_data else 'uri'
        if blob_type in ["s3", "azure"]:
            try:
                url = self.stored_data["presigned_url"]
            except KeyError:
                url = self.stored_data[key]
            if url.startswith("s3://"):
                url = self.stored_data["endpoint_url"] + "/" + url[5:]
            return url
        else:
            return self.stored_data[key]

    def download(self, filename=None, cache=True, head=None, progress_tracker=None):
        """Return a local filepath to the file this result points to."""
        if not filename:
            self._temp_filename = True
            myfile = NamedTemporaryFile(delete=False)
            myfile.close()
            filename = myfile.name
        blob_type = self.stored_data.get("__type__", "").lower()
        if cache and self._cached_filename:
            return self._cached_filename
        url = self.get_download_url()
        filepath = download_url(
            url, blob_type, filename,
            head=head, progress_tracker=progress_tracker
        )
        if cache:
            self._cached_filename = filepath
        return filepath
